#ifndef _ORDER_FILTER_H_
#define _ORDER_FILTER_H_

#include "Image.h"
#include "RunMedPHeap.h"
#include "RunMedHisto.h"

/*
 * This class is an implementation of an order filter and several median filters.
 * An order filter is a non linear filter which replaces the value of a pixel with
 * the value of the k'th smallest gray level value found in a window around the original
 * pixel. The median filter is simply k = 0.5*window-data.
 *
 * The returned image has the following structure:
 *                      -----------------------------------
 *                      |//////////////////////////////////|
 *                      |//|----------------------------|//|
 *                      |//|                            |//|
 *                      |//|                            |//|
 *                      |//|                            |//|
 *                      |//|     Filtered Data          |//|
 *                      |//|                            |//|
 *                      |//|                            |//|
 *                      |//|                            |//|
 *                      |//_____________________________ //|
 *                      |///////Original Data//////////////|
 *                      -----------------------------------
 * The outer frame has sizes according to the filter width and height.
 *                      
 * Author: Ziv Yaniv (zivy@cs.huji.ac.il)
 */

class OrderFilter {

public:
  /*
	 * sort the data found in the window and choose the
   * median for each pixel.
	 */
  static Image * naiveMedian(Image &img, 
                                int halfWidth, int halfHeight);
  /*
   * A selection filter which sets the value of a pixel
   * to the K'th smallest value in the window around it
   * according to the percentage.
   * See "Introduction to Algorithms", Cormen, Leiserson, Rivest, pp. 185-194
	 */
  static Image *hoareSelectK(Image &img, float percentage,
                                int halfWidth, int halfHeight);																
	/*
	 * Use the previous selection filter to find the median.
	 */
  static Image *hoareMedian(Image &img,
                                int halfWidth, int halfHeight) {
    return hoareSelectK(img,0.5,halfWidth,halfHeight);
  }

	/*
   * A selection filter which sets the value of a pixel
   * to the K'th smallest value in the window around it
   * according to the percentage. Uses a histogram, similar to
   * the running median with histogram.
	 * Use only if the data in the image is integral.
	 */
  static Image *histoSelectK(Image &img, float percentage,
                                int halfWidth, int halfHeight,
																int minGrayLevel, int maxGrayLevel);


	/*
	 * Apply a running median using partition heaps.
	 * See the RunMedPHeap.h file for documentation.
	 */
	static Image *RunningMedianReal(Image &img,
		                                 int halfWidth, int halfHeight) {
		RunMedPHeap *medFilter = new RunMedPHeap(halfWidth, halfHeight); 
		Image *res = medFilter->apply(img);
		delete medFilter;
		return(res);
	}

	/*
	 * Apply a running median using a histogram.
	 * See the RunMedHisto.h file for documentation.
	 */
	static Image *RunningMedianIntegral(Image &img,
		                                     int halfWidth, int halfHeight,
																				 int minGrayLevel, int maxGrayLevel) {
		RunMedHisto *medFilter = new RunMedHisto(halfWidth, halfHeight, 
			                                       minGrayLevel, maxGrayLevel); 
		Image *res = medFilter->apply(img);
		delete medFilter;
		return(res);
	}


	/*
	 * 3X3 is from XILINX XCELL magazine, vol. 23 by John L. Smith
   * http://www.xilinx.com/xcell/xl23/xl23_16.pdf
	 *
	 * NOTE: The median is computed leaving a border of one pixel.
	 *----
	 */
	static Image *median3X3(Image &img);

	/*
	 * 5X5 is from Graphics Gems pp.171-175 (Median finding on a 3X3 grid).
	 * The 5X5 appears only in the implementation section.
	 *
	 *NOTE: The median is computed leaving a border of two pixels.
	 *----
	 */
	static Image *median5X5(Image &img);

private:
            //returns the k'th smallest element found in the array data
            //Hoare's selection algorithm
  static float hoareSelect(float *data, int start, int end, int k);
	static inline void pixSort(float &pix1, float &pix2) {
		if(pix1>pix2) {
			float tmp = pix1;
			pix1 = pix2;
			pix2 = tmp;
		}
	}

	       //return the median of 3 elements
	static float median3(float *data);
	       //return the median of 5 elements
	static float median5(float *data);
	       //return the median of 7 elements
	static float median7(float *data);
	       //return the median of 9 elements
	static float median9(float *data);
			  //return the median of 25 elements
	static float median25(float *data);

};


#endif //_ORDER_FILTER_H_
