#include "RunMedHisto.h"

/*
 * Construct a new running median object. To enable easy access to the _histo
 * array according to the values found in the image I move the pointer back or
 * forward so that the first cell of the array is accessed by the call: 
 * _histo[_minGrayLevel].
 */
RunMedHisto::RunMedHisto(int halfWidth, int halfHeight, 
												int minGrayLevel, int maxGrayLevel) : 
                        _minGrayLevel(minGrayLevel),
												_maxGrayLevel(maxGrayLevel),
												_halfWidth(halfWidth),
	                      _halfHeight(halfHeight)
{
	_medianNumber = ((2*_halfWidth+1)*(2*_halfHeight+1))/2 + 1;

	_histo = new int[_maxGrayLevel - _minGrayLevel + 1];
	memset(_histo,'\0',(_maxGrayLevel - _minGrayLevel + 1)*sizeof(int));
	_histo -= _minGrayLevel*sizeof(int);
}
/*****************************************************************************/
/*
 * Move the pointer _histo to its original location and then free the 
 * allocated memory.
 */
RunMedHisto::~RunMedHisto()
{
	_histo+= _minGrayLevel*sizeof(int);
	delete [] _histo;
}
/*****************************************************************************/
/*
 * Traverse the image from top left to bottom right (follow the arrows).
 *
 *                 -----------------------
 *                 | --------->---------- |
 *                 |                    \/|
 *                 | ---------<----------<|
 *                 | \/                   |
 *                 |  >------------------ |
 *                 |                    \/|
 *                 |           :          |
 *                 |           :          |
 *                 |           :          |
 *                 ------------------------
 */
Image *RunMedHisto::apply(Image &img)
{
	int x, y, imgWidth, imgHeight, minX, maxX, minY, maxY;
	int leavingX, enteringX, startY, endY, curY;
	bool leftToRight;

	imgWidth = img.getWidth();
	imgHeight = img.getHeight();

	Image *res = new Image(img);

	minY = _halfHeight;
	maxY = imgHeight - _halfHeight - 1;
	minX = _halfWidth;
  maxX = imgWidth-1-_halfWidth;
	leftToRight = true;
	
	initData(img);  //initialize the first histogram (top left)
	res->Pix(_halfWidth,_halfHeight) = getMedian();
	
	for(y=minY; y<=maxY; y++) {
		startY = y - _halfHeight;
		endY = y + _halfHeight;
		if(leftToRight) { //traverse row from left to right
			for(x=minX+1; x<=maxX; x++) {
				leavingX = x - _halfWidth - 1;
				enteringX = x + _halfWidth;
				for(curY = startY; curY <= endY; curY++) {
					_histo[(int)img.Pix(leavingX,curY)]--;
					_histo[(int)img.Pix(enteringX,curY)]++;
				}
				res->Pix(x,y) = getMedian();
			}
			x--;
		}
		else {          //traverse row from right to left
			for(x=maxX-1; x>=minX; x--) {
				leavingX = x + _halfWidth + 1;
				enteringX = x - _halfWidth;
				for(curY = startY; curY <= endY; curY++) {
					_histo[(int)img.Pix(leavingX,curY)]--;
					_histo[(int)img.Pix(enteringX,curY)]++;
				}
				res->Pix(x,y) = getMedian();
			}
			x++;
		}
		if(leftToRight)
			leftToRight = false;
		else 
			leftToRight = true;
		if(y+1 <= maxY) {  //initialize next row
			initNextRow(img,x,y);
			res->Pix(x,y+1) = getMedian();
		}
	}
	return(res);
}
/*****************************************************************************/
/*
 * Initialize the histogram to contain the data found in the top left corner
 * of the image.
 */
void RunMedHisto::initData(Image &img)
{
	int width = 2*_halfWidth+1;
	int height = 2*_halfHeight+1;

	for(int y = 0; y<height; y++) 
		for(int x = 0; x<width; x++) 
			_histo[(int)img.Pix(x,y)]++;
}
/*****************************************************************************/
/*
 * Update the histogram for a new row.
 */
void RunMedHisto::initNextRow(Image &img, int x, int y)
{
	int leavingY, enteringY, startX, endX, curX;

	leavingY = y - _halfHeight;
	enteringY = y + _halfHeight + 1;
	startX = x - _halfWidth;
	endX = x + _halfWidth;

	for(curX = startX; curX<=endX; curX++) {
		_histo[(int)img.Pix(curX,leavingY)]--;
		_histo[(int)img.Pix(curX,enteringY)]++;
	}
}
/*****************************************************************************/
/*
 * Return the median of the _histo data.
 */
int RunMedHisto::getMedian()
{
	int i, curNumber;

	i= _minGrayLevel;
	curNumber = _histo[i];
	while(curNumber < _medianNumber) {
		i++;
		curNumber+= _histo[i];
	}
	return(i);
}
