#include <stdlib.h>
#include <time.h>
#include "RunMedPHeap.h"


/*
 * Constructor receives the size of the median window which will be:
 * (2*halfWidth+1) X (2*halfHeight+1)
 */
RunMedPHeap::RunMedPHeap(int halfWidth, int halfHeight) : _halfWidth(halfWidth),
	                                              _halfHeight(halfHeight)
{
	int i;
	int width = 2*_halfWidth+1;
	int height = 2*_halfHeight+1;

	_window = new float*[width];
	for(i=0; i<width; i++)
		_window[i] = new float[height];

	_window2Heap = new int*[width];
	for(i=0; i<width; i++)
		_window2Heap[i] = new int[height];

	_partitionHeap = new Pnt[width*height];

  _medianIndex = ((2*_halfWidth+1)*(2*_halfHeight+1))/2;
	
	_maxHeap = _partitionHeap;
	_minHeap = &(_partitionHeap[_medianIndex+1]);
}
/*****************************************************************************/
RunMedPHeap::~RunMedPHeap()
{

	int i;
	int width = 2*_halfWidth+1;
	int height = 2*_halfHeight+1;

	for(i=0; i<width; i++)
		delete [] _window[i];
	delete [] _window;

	for(i=0; i<width; i++)
		delete [] _window2Heap[i];
	delete [] _window2Heap;

	delete [] _partitionHeap;
}
/*****************************************************************************/
/*
 * Traverse the image from left to right (follow the arrows).
 *
 *                 -----------------------
 *                 | --------->---------- |
 *                 | --------->---------- |
 *                 | --------->---------- |
 *                 |          :           |
 *                 |          :           |
 *                 ------------------------
 */
Image *RunMedPHeap::apply(Image &img)

{

	int x, y, curY, imgWidth, imgHeight, minX, maxX;
	int windowStartX, windowEndX, windowX, windowY;

	imgWidth = img.getWidth();
	imgHeight = img.getHeight();

	Image *res = new Image(img);

  minX = _halfWidth;
  maxX = imgWidth-1-_halfWidth;
  windowStartX = 0;
  windowEndX = 2*_halfWidth;
	for(y=_halfHeight; y<imgHeight-_halfHeight; y++) {
		initData(img,y,minX);
		res->Pix(minX,y) = _window[_partitionHeap[_medianIndex].x][_partitionHeap[_medianIndex].y];
		windowX = 0;
		for(x=minX+1; x<=maxX; x++) {
			for(curY=y-_halfHeight,windowY=0; windowY<2*_halfHeight+1; curY++,windowY++) {
				updateData(windowX,windowY,img.Pix(x+_halfWidth,curY));
			}
			windowX++;
      if(windowX > windowEndX)
				windowX = windowStartX;
			res->Pix(x,y) = _window[_partitionHeap[_medianIndex].x][_partitionHeap[_medianIndex].y];
    }
  }
	return(res);
}
/*****************************************************************************/
/*
 * Initialize the data structures, mainly the partition heap.
 * Root of the max heap is found at _partitionHeap[0] and root of the min
 * heap is at _partitionHeap[_medianIndex+1].
 */
void RunMedPHeap::initData(Image &img, int y, int x)
{
	int windowY,windowX,imgX,imgY;
	int windowWidth = 2*_halfWidth+1;
	int windowHeight = 2*_halfHeight+1;
	
	for(windowY=0, imgY= y-_halfHeight; windowY<windowHeight; windowY++, imgY++) {
		for(windowX=0, imgX= x-_halfWidth; windowX<windowWidth; windowX++, imgX++) {
			_window[windowX][windowY] = img.Pix(imgX,imgY);
      _window2Heap[windowX][windowY] = windowY*windowWidth + windowX;
      _partitionHeap[_window2Heap[windowX][windowY]].x = windowX;
      _partitionHeap[_window2Heap[windowX][windowY]].y = windowY;
    }
  }
	
	quickSort(0,windowWidth*windowHeight-1);

  int sz = (windowWidth*windowHeight)/2;
  Pnt *tmp = new Pnt[sz];
  memcpy(tmp,_partitionHeap,sz*sizeof(Pnt));
  for(int i=0; i<sz; i++) {
    _partitionHeap[i] = tmp[sz-1-i]; 
		_window2Heap[_partitionHeap[i].x][_partitionHeap[i].y] = i;
	}
  delete [] tmp;    
}
/*****************************************************************************/
void RunMedPHeap::swap(Pnt &p1, Pnt &p2)
{
  Pnt tmpPnt;
  int tmpIndex;
  
  tmpIndex = _window2Heap[p1.x][p1.y];
  _window2Heap[p1.x][p1.y] = _window2Heap[p2.x][p2.y];
  _window2Heap[p2.x][p2.y] = tmpIndex;

  tmpPnt = p1;
  p1 = p2;
  p2 = tmpPnt;
}
/*****************************************************************************/
/*
 * Randomized partiton used by the quickSort method to perform the 
 * actual sorting.
 */
int RunMedPHeap::partition(int start, int end)
{
  float pivotValue;
  int pivotIndex, left, right;

      //seed the random number generator
  srand((unsigned)time(NULL));
            //select a random pivot
  pivotIndex = (int)(rand() * ((float)(end-start)/(float)RAND_MAX) + start);
  pivotValue = _window[_partitionHeap[pivotIndex].x][_partitionHeap[pivotIndex].y];

  swap(_partitionHeap[start],_partitionHeap[pivotIndex]);

  left = start+1;
  right = end;

  while(left <= right) {
    while((left <= end) && (_window[_partitionHeap[left].x][_partitionHeap[left].y] <= pivotValue))
      left++;
    while(_window[_partitionHeap[right].x][_partitionHeap[right].y] > pivotValue)
      right--;
    if(left < right)
      swap(_partitionHeap[left],_partitionHeap[right]);
  }
  swap(_partitionHeap[start],_partitionHeap[right]);

	return(right);
}
/*****************************************************************************/
/*
 * Sort the elements of the _partitionHeap according to the values found in
 * _window. 
 * NOTE: As quickSort is recursive if the sorted array _window is large or
 *       the partitions are uneven stack depth may cause trouble.
 */
void RunMedPHeap::quickSort(int start, int end)
{ 
  int sz = end - start;
  int pivot;
	
  if(sz > 0) {
    if(sz == 1) {    //two elements, don't recurse
      if(_window[_partitionHeap[start].x][_partitionHeap[start].y] < 
        _window[_partitionHeap[end].x][_partitionHeap[end].y])
        swap(_partitionHeap[start],_partitionHeap[end]);
    }
    else if(sz == 2) {  //three elements, don't recurse
      if(_window[_partitionHeap[start].x][_partitionHeap[start].y] >
				 _window[_partitionHeap[start+1].x][_partitionHeap[start+1].y])
        swap(_partitionHeap[start],_partitionHeap[start+1]);
      if(_window[_partitionHeap[start+1].x][_partitionHeap[start+1].y] > 
				 _window[_partitionHeap[end].x][_partitionHeap[end].y])
				swap(_partitionHeap[start+1],_partitionHeap[end]);
      if(_window[_partitionHeap[start].x][_partitionHeap[start].y] >
				 _window[_partitionHeap[start+1].x][_partitionHeap[start+1].y])
        swap(_partitionHeap[start],_partitionHeap[start+1]);
    }
    else {                //more than three elements
      pivot = partition(start,end);
			//push the larger sub problem onto the stack first
      if(end-pivot > pivot-start) { 
        quickSort(pivot+1,end);
        quickSort(start,pivot-1);
      }
      else {
        quickSort(start,pivot-1);
        quickSort(pivot+1,end);
      }
    }
  }
}
/*****************************************************************************/
/*
 * Insert the value 'val' into the partition heaps replacing the element
 * found in _window[windowX][windowY].
 */
void RunMedPHeap::updateData(int windowX, int windowY, float val)
{
	_window[windowX][windowY] = val;

	            //element that leaves is from the upper heap
	if(_window2Heap[windowX][windowY] > _medianIndex) {
		if(_window[windowX][windowY] >= _window[_partitionHeap[_medianIndex].x][_partitionHeap[_medianIndex].y])
		  minOutMinIn(windowX,windowY);
		else
		  minOutMaxIn(windowX,windowY);
	}             //element that leaves is from the lower heap
	else if(_window2Heap[windowX][windowY] < _medianIndex) {
		if(_window[windowX][windowY] < _window[_partitionHeap[_medianIndex].x][_partitionHeap[_medianIndex].y])
		  maxOutMaxIn(windowX,windowY);
		else
		  maxOutMinIn(windowX,windowY);
	}       //element that leaves is the median
	else {
		if(_window[_partitionHeap[_medianIndex].x][_partitionHeap[_medianIndex].y] > 
			_window[_partitionHeap[_medianIndex+1].x][_partitionHeap[_medianIndex+1].y])
		  medianOutMinIn(windowX,windowY);
		else if (_window[_partitionHeap[_medianIndex].x][_partitionHeap[_medianIndex].y] <
			       _window[_partitionHeap[0].x][_partitionHeap[0].y])                           
			medianOutMaxIn(windowX,windowY);
		//else medianOutMedianIn(), don't do anything
	}
}
/*****************************************************************************/
/*
 * The new value and the removed value both come from the min heap.
 * The new value is placed in the heap instead of the old value and the
 * heap is modified to maintain its invariants.
 * To maintain the min heap structure:
 *                                    1.full binary tree
 *                                    2.parent < leftChild and 
 *                                      parent < rightChild, (left>right or
 *                                      left<right doesn't matter.
 * we try to (a) push the new value towards the leaves as far as possible and
 * then (b) push it as far as possible towards the root.
 */
void RunMedPHeap::minOutMinIn(int windowX, int windowY)
{
  int cur, lChild, rChild, minIndex, parent;
	int heapSize = _medianIndex;

             // (a) push the new value down the min heap (towards the leaves) 
  cur = _window2Heap[windowX][windowY] - _medianIndex - 1;
	lChild = 2*cur+1; 
	rChild = lChild+1;
	if(rChild < heapSize)       //two children
		minIndex = _window[_minHeap[lChild].x][_minHeap[lChild].y] > 
				       _window[_minHeap[rChild].x][_minHeap[rChild].y] ? rChild : lChild;
	else if(lChild < heapSize) //one child
		minIndex = lChild;
	else 
		minIndex = heapSize;  //no children (a leaf)
	while((minIndex < heapSize) &&
		    (_window[_minHeap[cur].x][_minHeap[cur].y] >
				 _window[_minHeap[minIndex].x][_minHeap[minIndex].y])) {
    swap(_minHeap[cur],_minHeap[minIndex]);
    cur = minIndex;
    lChild = 2*cur+1; 
    rChild = lChild+1;
		if(rChild < heapSize)       //two children
			minIndex = _window[_minHeap[lChild].x][_minHeap[lChild].y] > 
					       _window[_minHeap[rChild].x][_minHeap[rChild].y] ? rChild : lChild;
		else if(lChild < heapSize) //one child
			minIndex = lChild;
		else 
			minIndex = heapSize;  //no children (a leaf)
  }
                    //(b) push the new value up the min heap (towards the root) 
  parent = (cur-1)/2;
  while(_window[_minHeap[cur].x][_minHeap[cur].y] <
        _window[_minHeap[parent].x][_minHeap[parent].y]) {
    swap(_minHeap[cur],_minHeap[parent]);
    cur = parent;
    parent = (cur-1)/2;
  }                
}
/*****************************************************************************/
/*
 * The removed value comes from the min heap and the new value is inserted
 * into the max heap.
 * To maintain both heap structures we (a) push the hole in the min heap
 * to the root and then (b) set its root to be the previous median.
 * (c) Then we push the new value from the median position down the 
 * max heap (towards the leaves) as far as possible.
 */
void RunMedPHeap::minOutMaxIn(int windowX, int windowY)
{
	int cur, parent, lChild, rChild, maxIndex;
	int heapSize = _medianIndex;
                  //(a) push the hole to the root of the min heap
	cur = _window2Heap[windowX][windowY] - _medianIndex - 1;
  while(cur != 0) {
		parent = (cur-1)/2;
		swap(_minHeap[cur],_minHeap[parent]);
		cur = parent;
	}

	      //(b) new value is now in the median's place
	swap(_partitionHeap[_medianIndex+1],_partitionHeap[_medianIndex]);
	cur = _medianIndex;

	               //(c) push the new value down the max heap
	if(_window[_partitionHeap[cur].x][_partitionHeap[cur].y] < 
		 _window[_partitionHeap[0].x][_partitionHeap[0].y]) {
		swap(_partitionHeap[cur],_partitionHeap[0]);
		cur = 0;
		lChild = 2*cur+1; 
		rChild = lChild+1;

		if(rChild < heapSize)       //two children
			maxIndex = _window[_maxHeap[lChild].x][_maxHeap[lChild].y] > 
				         _window[_maxHeap[rChild].x][_maxHeap[rChild].y] ? lChild : rChild;
		else if(lChild < heapSize) //one child
			maxIndex = lChild;
		else 
			maxIndex = heapSize;  //no children (a leaf)

		while((maxIndex < heapSize) &&
			    (_window[_maxHeap[cur].x][_maxHeap[cur].y] <
					 _window[_maxHeap[maxIndex].x][_maxHeap[maxIndex].y])) {
			swap(_maxHeap[cur],_maxHeap[maxIndex]);
			cur = maxIndex;
			lChild = 2*cur+1; 
			rChild = lChild+1;
			if(rChild < heapSize)       //two children
				maxIndex = _window[_maxHeap[lChild].x][_maxHeap[lChild].y] > 
					         _window[_maxHeap[rChild].x][_maxHeap[rChild].y] ? lChild : rChild;
			else if(lChild < heapSize) //one child
				maxIndex = lChild;
			else 
				maxIndex = heapSize;  //no children (a leaf)
		}
	}
}
/*****************************************************************************/
/*
 * The new value and the removed value both come from the max heap.
 * The new value is placed in the heap instead of the old value and the
 * heap is modified to maintain its invariants.
 * To maintain the max heap structure:
 *                                    1.full binary tree
 *                                    2.parent > leftChild and 
 *                                      parent > rightChild, (left>right or
 *                                      left<right doesn't matter.
 * we try to (a) push the new value towards the leaves as far as possible and
 * then (b) push it as far as possible towards the root.
 */
void RunMedPHeap::maxOutMaxIn(int windowX, int windowY)
{
  int cur, lChild, rChild, maxIndex, parent;
	int heapSize = _medianIndex;
             //(a) push the new value down the max heap (towards the leaves) 
  cur = _window2Heap[windowX][windowY];
  lChild = 2*cur+1; 
  rChild = lChild+1;

	if(rChild < heapSize)       //two children
	  maxIndex = _window[_maxHeap[lChild].x][_maxHeap[lChild].y] > 
		           _window[_maxHeap[rChild].x][_maxHeap[rChild].y] ? lChild : rChild;
	else if(lChild < heapSize) //one child
		maxIndex = lChild;
	else 
		maxIndex = heapSize;  //no children (a leaf)

  while((maxIndex < heapSize) &&
		    (_window[_maxHeap[cur].x][_maxHeap[cur].y] <
				 _window[_maxHeap[maxIndex].x][_maxHeap[maxIndex].y])) {
    swap(_maxHeap[cur],_maxHeap[maxIndex]);
    cur = maxIndex;
    lChild = 2*cur+1; 
    rChild = lChild+1;
		if(rChild < heapSize)       //two children
		  maxIndex = _window[_maxHeap[lChild].x][_maxHeap[lChild].y] > 
			           _window[_maxHeap[rChild].x][_maxHeap[rChild].y] ? lChild : rChild;
		else if(lChild < heapSize) //one child
			maxIndex = lChild;
		else 
			maxIndex = heapSize;  //no children (a leaf)
  }
                    //push the new value up the max heap (towards the root) 
  parent = (cur-1)/2;
  while(_window[_maxHeap[cur].x][_maxHeap[cur].y] >
        _window[_maxHeap[parent].x][_maxHeap[parent].y]) {
    swap(_maxHeap[cur],_maxHeap[parent]);
    cur = parent;
    parent = (cur-1)/2;
  } 
	
}
/*****************************************************************************/
/*
 * The removed value comes from the max heap and the new value is inserted
 * into the min heap.
 * To maintain both heap structures we (a) push the hole in the max heap
 * to the root and then (b) set its root to be the previous median.
 * (c) Then we push the new value from the median position down the 
 * min heap (towards the leaves) as far as possible.
 */
void RunMedPHeap::maxOutMinIn(int windowX, int windowY)
{
	int cur, parent, lChild, rChild, minIndex;
	int heapSize = _medianIndex;

                  // (a) push the hole to the root of the max heap
	cur = _window2Heap[windowX][windowY];
  while(cur != 0) {
		parent = (cur-1)/2;
		swap(_maxHeap[cur],_maxHeap[parent]);
		cur = parent;
	}

	      // (b) new value is now in the median's place
	swap(_partitionHeap[cur],_partitionHeap[_medianIndex]);
	cur = _medianIndex;

		               //(c) push the new value down the min heap
	if(_window[_partitionHeap[cur].x][_partitionHeap[cur].y] > 
		_window[_partitionHeap[_medianIndex+1].x][_partitionHeap[_medianIndex+1].y]) {
		swap(_partitionHeap[cur],_partitionHeap[_medianIndex+1]);

		cur = 0;
		lChild = 2*cur+1; 
		rChild = lChild+1;
		if(rChild < heapSize)       //two children
			minIndex = _window[_minHeap[lChild].x][_minHeap[lChild].y] > 
					       _window[_minHeap[rChild].x][_minHeap[rChild].y] ? rChild : lChild;
		else if(lChild < heapSize) //one child
			minIndex = lChild;
		else 
			minIndex = heapSize;  //no children (a leaf)
		while((minIndex < heapSize) &&
			    (_window[_minHeap[cur].x][_minHeap[cur].y] >
					 _window[_minHeap[minIndex].x][_minHeap[minIndex].y])) {
			swap(_minHeap[cur],_minHeap[minIndex]);
			cur = minIndex;
			lChild = 2*cur+1; 
			rChild = lChild+1;
			if(rChild < heapSize)       //two children
				minIndex = _window[_minHeap[lChild].x][_minHeap[lChild].y] > 
						       _window[_minHeap[rChild].x][_minHeap[rChild].y] ? rChild : lChild;
			else if(lChild < heapSize) //one child
				minIndex = lChild;
			else 
				minIndex = heapSize;  //no children (a leaf)
		}
	}
}
/*****************************************************************************/
/*
 * The removed value is the median and the new value is inserted
 * into the min heap.
 * To maintain the min heap structure we (a) swap the new value from the median
 * position and the root of the min heap. (b) Push the new value down the 
 * min heap (towards the leaves) as far as possible.
 */
void RunMedPHeap::medianOutMinIn(int windowX, int windowY)
{ 
	int lChild, rChild, minIndex, cur;
	int heapSize = _medianIndex;

	        // (a) swap the new value and the min heap root
	swap(_partitionHeap[_medianIndex],_partitionHeap[_medianIndex+1]);

	       // (b) push the new value down the min heap (towards the leaves) 
	cur = 0;
	lChild = 2*cur+1; 
	rChild = lChild+1;
	if(rChild < heapSize)       //two children
		minIndex = _window[_minHeap[lChild].x][_minHeap[lChild].y] > 
				       _window[_minHeap[rChild].x][_minHeap[rChild].y] ? rChild : lChild;
	else if(lChild < heapSize) //one child
		minIndex = lChild;
	else 
		minIndex = heapSize;  //no children (a leaf)
	while((minIndex < heapSize) &&
		    (_window[_minHeap[cur].x][_minHeap[cur].y] >
				 _window[_minHeap[minIndex].x][_minHeap[minIndex].y])) {
    swap(_minHeap[cur],_minHeap[minIndex]);
    cur = minIndex;
    lChild = 2*cur+1; 
    rChild = lChild+1;
		if(rChild < heapSize)       //two children
			minIndex = _window[_minHeap[lChild].x][_minHeap[lChild].y] > 
					       _window[_minHeap[rChild].x][_minHeap[rChild].y] ? rChild : lChild;
		else if(lChild < heapSize) //one child
			minIndex = lChild;
		else 
			minIndex = heapSize;  //no children (a leaf)
  }

}
/*****************************************************************************/
/*
 * The removed value is the median and the new value is inserted
 * into the max heap.
 * To maintain the max heap structure we (a) swap the new value from the median
 * position and the root of the max heap. (b) Push the new value down the 
 * max heap (towards the leaves) as far as possible.
 */
void RunMedPHeap::medianOutMaxIn(int windowX, int windowY)
{
	int lChild, rChild, maxIndex, cur;
	int heapSize = _medianIndex;

      	// (a) swap the new value and the max heap root
	swap(_partitionHeap[_medianIndex],_partitionHeap[0]);

	      // (b) push the new value down the max heap (towards the leaves) 
  cur = 0;
  lChild = 2*cur+1; 
  rChild = lChild+1;

	if(rChild < heapSize)       //two children
	  maxIndex = _window[_maxHeap[lChild].x][_maxHeap[lChild].y] > 
		           _window[_maxHeap[rChild].x][_maxHeap[rChild].y] ? lChild : rChild;
	else if(lChild < heapSize) //one child
		maxIndex = lChild;
	else 
		maxIndex = heapSize;  //no children (a leaf)

  while((maxIndex < heapSize) &&
		    (_window[_maxHeap[cur].x][_maxHeap[cur].y] <
				 _window[_maxHeap[maxIndex].x][_maxHeap[maxIndex].y])) {
    swap(_maxHeap[cur],_maxHeap[maxIndex]);
    cur = maxIndex;
    lChild = 2*cur+1; 
    rChild = lChild+1;
		if(rChild < heapSize)       //two children
		  maxIndex = _window[_maxHeap[lChild].x][_maxHeap[lChild].y] > 
			           _window[_maxHeap[rChild].x][_maxHeap[rChild].y] ? lChild : rChild;
		else if(lChild < heapSize) //one child
			maxIndex = lChild;
		else 
			maxIndex = heapSize;  //no children (a leaf)
  }

}

