#ifndef _RUN_MED_P_HEAP_H_
#define _RUN_MED_P_HEAP_H_

#include "Image.h"


/*
 * Implementation of the running median from the article:
 * "Optimal Median Smoothing", W. Hardle, W. Steiger, 1994.
 *
 * NOTE: This median filter can work with continous values (i.e. real). 
 *------
 *
 * The data structures are organized as follows:
 *
 *         _______                         max heap <=  median <=  min heap           
 *         |     |                      _______________________________________
 * window: |     |     partition heaps: |                 | |                  |
 *         |     |                      ----------------------------------------
 *         |     |                       /\                  /\
 *         -------MxN                 max heap             min heap
 *                                      root                 root   
 *              _______
 *              |     |
 * window2Heap: |     |
 *              |     |
 *              |     |
 *              -------MxN
 *                   
 *
 *
 * window: the actual values we are dealing with.
 * partition heaps: A max and a min heap which contain indexes (x,y) into the 
 *                  window. The median is always located at 
 *                  ((2*_halfWidth+1)*(2*_halfHeight+1))/2 in the array.
 * window2Heap: given an index (x,y) in the window this array contains the index
 *              (i) of the corresponding element in the partition heaps array.
 *
 * Author: Ziv Yaniv (zivy@cs.huji.ac.il)
 */



class RunMedPHeap {

public:
    //median using a window of size 2*halfWidth+1 X 2*halfHeight+1 
	RunMedPHeap(int halfWidth, int halfHeight); 
	~RunMedPHeap();
   /*
    * Apply the median filter to the given image.
    * Freeing the memory used by the returned image is the users 
    * responsibility.
    */
	Image *apply(Image &img);

private:            

	typedef struct {
		int x;
		int y;
	} Pnt;

  void swap(Pnt &p1, Pnt &p2);
  int partition(int start, int end);
  void quickSort(int start, int end);
	void updateData(int windowX, int windowY, float val);
	void initData(Image &img, int y, int x);
	void minOutMinIn(int windowX, int windowY);
	void minOutMaxIn(int windowX, int windowY);
	void maxOutMaxIn(int windowX, int windowY);
	void maxOutMinIn(int windowX, int windowY);
	void medianOutMinIn(int windowX, int windowY);
	void medianOutMaxIn(int windowX, int windowY);

	int _halfWidth;
	int _halfHeight;
  int _medianIndex;
	float **_window;
	int **_window2Heap;
	Pnt *_partitionHeap;
	Pnt *_maxHeap;
	Pnt *_minHeap;
};



#endif //_RUN_MED_P_HEAP_H_
